from . import log
logger = log.get(__name__)

class Processor:
    
    def __init__(self, registry):
        self._registry = registry
        self._messageHandlers = {
            "policies": self._acceptPolicies,
            "endpoints": self._acceptEndpoints,
            "peers": self._acceptPeers,
            "consumes": self._acceptMetaConsumes
        }

    def accept(self, section, data):
        logger.debug('Accept Section: %s, data: %s', section, data)

        handler = self._messageHandlers.get(section)
        if handler is not None: 
            return handler(data)

    def _acceptPolicies(self, data):
        if data is not None:
            self._registry.set('policies', [], data)
        else:
            self._registry.reset('policies')

    def _acceptMetaConsumes(self, data):
        if data is not None:
            self._registry.set('consumes', [], data)
        else:
            self._registry.reset('consumes')

    def _acceptEndpoints(self, data):
        if data is not None:
            self._registry.set('endpoints', [], data)
            for endpointName, endpoints in data.items():
                self._registry.set('endpoints', [endpointName], endpoints)
        else:
            self._registry.reset('endpoints')

    def _acceptPeers(self, data):
        if data is None:
            self._registry.reset('peers')
        else:
            self._registry.set('peers', [], data)
            for serviceId, serviceData in data.items():
                self._handlePeers(serviceId, serviceData)
            
    def _handlePeers(self, serviceId, serviceData):
        self._registry.set('peers', [serviceId], serviceData)



from . import log
logger = log.get(__name__)

from .deep_eq import deep_eq
import random as rand

class PeerHelper:
    def __init__(self, registry):
        self._registry = registry

    def monitorPeers(self, peerPath, cb):
        self._registry.subscribe('peers', peerPath, cb)

    def monitorPeer(self, peerPath, selector, cb):
        class nonlocal:
            oldValue = None
        
        def innerCb(peers):
            value = selector(peers)
            isChanged = False
            if value is not None:
                if nonlocal.oldValue is not None:
                    isChanged = not deep_eq(value, nonlocal.oldValue)
                else:
                    isChanged = True
            else:
                if nonlocal.oldValue is not None:
                    isChanged = True
                else:
                    isChanged = False
            if isChanged:
                nonlocal.oldValue = value
                cb(value)

        self._registry.subscribe('peers', peerPath, innerCb)

    def getPeers(self, peerPath):
        return self._registry.get('peers', peerPath)

    def getPeer(self, peerPath, selector):
        peers = self._registry.get('peers', peerPath)
        return selector(peers)

    def selectFirstPeer(self, peers):
        return firstFromDict(peers)

    def selectRandomPeer(self, peers):
        return randomFromDict(peers)




def randomFromList(list):
    return rand.choice(list)

def firstFromList(list):
    return list[0]

def randomFromDict(dict):
    if not dict:
        return None
    key = randomFromList(list(dict.keys()))
    return dict[key]

def firstFromDict(dict):
    if not dict:
        return None
    key = firstFromList(list(dict.keys()))
    return dict[key]


from . import log
logger = log.get(__name__)

logger.info('Starting...')

import sys
import json
import random as rand

from .registry import Registry
from .policy import Policy
from .zipkin import Zipkin
from .processor import Processor
from .client import Client
from .executor import Executor
from .peer_helper import PeerHelper

import copy
import os

berlioz_cluster = os.environ.get('BERLIOZ_CLUSTER')
berlioz_sector = os.environ.get('BERLIOZ_SECTOR')
berlioz_service = os.environ.get('BERLIOZ_SERVICE')


def onMessage(msg):
    for section, data in msg.items():
        # logger.info('Section %s, data: %s', section, data)
        processor.accept(section, data)
    # logger.info('**** REGISTRY: %s', json.dumps(registry.extractRoot(), indent=4, sort_keys=True))


def instrument(method, binary_annotations):
    return zipkin.instrument(method, binary_annotations)


def setupFlask(app):
    from .frameworks.b_flask import Flask
    Flask(app, zipkin, policy, registry)


registry = Registry()
peerHelper = PeerHelper(registry)
policy = Policy(registry)
zipkin = Zipkin(peerHelper, policy)
processor = Processor(registry)
client = Client(onMessage)




from . import log
logger = log.get(__name__)

import requests
import json

from .peer_accessor import PeerAccessor

class HttpPeerAccessor(PeerAccessor):
    def __init__(self, starter, id, endpoint):
        if endpoint is None:
            endpoint = 'default'
        serviceId = id + '-' + endpoint
        PeerAccessor.__init__(self, starter, [serviceId])

    def request(self ):
        return RequestWrapper(self)

class RequestWrapper(object):

    def __init__(self, target):
        object.__setattr__(self, "_target", target)

    def __getattribute__(self, propKey):

        def perform(*args, **kwargs):
            print(args)
            print(kwargs)
            kwargs.setdefault('headers', {})

            target = object.__getattribute__(self, "_target")
            if propKey == 'request':
                method = args[0]
                url = args[1]
            else:
                method = propKey 
                url = args[0]

            def execAction(peer, zipkin_span=None):
                print(peer)
                print(url)
                origMethod = getattr(requests, propKey)
                newargs = []
                if propKey == 'request':
                    newargs.append(method)    
                finalUrl = peer['protocol'] + '://' + peer['address'] + ':' + str(peer['port'])
                if url:
                    finalUrl = finalUrl + url
                newargs.append(finalUrl)
                if zipkin_span:
                    target._starter.zipkin.addZipkinHeaders(kwargs['headers'], zipkin_span)
                result = origMethod(*newargs, **kwargs)
                return result
            binary_annotations = {
                'http.url': url
            }
            return target.performExecutor(method, binary_annotations, execAction)

        return perform




from . import log
logger = log.get(__name__)

import json


class Policy:
    
    def __init__(self, registry):
        self._registry = registry
        self._defaults = {
            'enable-zipkin': False,
            'zipkin-service-id': '',
            'timeout': 5000,
            'no-peer-retry': True,
            'retry-count': 3,
            'retry-initial-delay': 500,
            'retry-delay-multiplier': 2,
            'retry-max-delay': 5000
        }

    def monitor(self, name, target, cb):
        context = {
            "value": self.resolve(name, target),
        }
        cb(context["value"])
        self._registry.subscribe('policies', [], lambda l: self._processMonitor(context, name, target, cb))

    def _processMonitor(self, context, name, target, cb):
        newValue = self.resolve(name, target)
        if context["value"] != newValue:
            context["value"] = newValue
            cb(newValue)

    def resolve(self, name, target):
        root = self._registry.get('policies', [])
        if root is None:
            root = {}
        if target is None:
            target = []

        value = self._resolve(root, name, target)
        if value is not None:
            return value

        value = self._defaults.get(name)
        if value is not None:
            return value
        
        logger.error('No Default set for %s', name)
        return None

    def _resolve(self, root, name, target):
        value = None
        if target:
            children = root.get('children')
            if children:
                child = children.get(target[0])
                if child:
                    childTarget = target[1:]
                    value = self._resolve(child, name, childTarget)
        if value:
            return value
        values = root.get('values')
        if values:
            return values.get(name)
        return None


from . import log
logger = log.get(__name__)

from .http_peer_accessor import HttpPeerAccessor

class Service(HttpPeerAccessor):
    def __init__(self, starter, id, endpoint):
        HttpPeerAccessor.__init__(self, starter, id, endpoint)




import boto3

def fetchDynamoClient(peer):
    kwargs = _getClientConfig(peer)
    dynamodb = boto3.resource('dynamodb', **kwargs)
    return dynamodb.Table(peer['name'])

def fetchKinesisClient(peer):
    kwargs = _getClientConfig(peer)
    return boto3.client('kinesis', **kwargs)

def fetchSSMClient(peer):
    kwargs = _getClientConfig(peer)
    return boto3.client('ssm', **kwargs)

def setupKinesisArgs(peer, actionName, xargs, xkwargs):
    if actionName == 'get_records':
        return
    if actionName == 'list_shards' and 'NextToken' in xkwargs:
        return
    xkwargs['StreamName'] = peer['name']

def setupParameterArgs(peer, actionName, xargs, xkwargs):
    xkwargs['Name'] = peer['name']

def _getClientConfig(peer):
    kwargs = {}
    peerConfig = peer.get('config')
    if peerConfig:
        region = peerConfig.get('region')
        if region:
            kwargs['region_name'] = region
        credentials = peerConfig.get('credentials')
        if credentials:
            key = credentials.get('accessKeyId')
            if key:
                kwargs['aws_access_key_id'] = key
            secret = credentials.get('secretAccessKey')
            if secret:
                kwargs['aws_secret_access_key'] = secret
    return kwargs